use cgmath::Point3;
use wgpu::util::DeviceExt;

#[rustfmt::skip]
pub const OPENGL_TO_WGPU_MATRIX: cgmath::Matrix4<f32> = cgmath::Matrix4::new(
    1.0, 0.0, 0.0, 0.0,
    0.0, 1.0, 0.0, 0.0,
    0.0, 0.0, 0.5, 0.5,
    0.0, 0.0, 0.0, 1.0,
);

pub const DEFAULT_CAMERA_ZOOM: f32 = 1.7;

pub struct Camera {
	pub eye: cgmath::Point3<f32>,
	pub target: cgmath::Point3<f32>,
	pub up: cgmath::Vector3<f32>,
	fovy: f32,
	znear: f32,
	zfar: f32,
}

pub struct CameraInfo {
	pub uniform: CameraUniform,
	pub buffer: wgpu::Buffer,
	pub layout: wgpu::BindGroupLayout,
	pub bind_group: wgpu::BindGroup,
}

impl Camera {
	pub fn new(device: &wgpu::Device) -> (Self, CameraInfo) {
		let camera = Self {
			eye: (0.0, 0.0, DEFAULT_CAMERA_ZOOM).into(),
			target: (0.0, 0.0, 0.0).into(),
			up: cgmath::Vector3::unit_y(),
			fovy: 45.0,
			znear: 0.1,
			zfar: 100.0,
		};

		let layout = device.create_bind_group_layout(&wgpu::BindGroupLayoutDescriptor {
			entries: &[wgpu::BindGroupLayoutEntry {
				binding: 0,
				visibility: wgpu::ShaderStages::VERTEX,
				ty: wgpu::BindingType::Buffer {
					ty: wgpu::BufferBindingType::Uniform,
					has_dynamic_offset: false,
					min_binding_size: None,
				},
				count: None,
			}],
			label: Some("camera_bind_group_layout"),
		});

		let mut uniform = CameraUniform::new();
		uniform.update_view_proj(&camera);

		let buffer = device.create_buffer_init(&wgpu::util::BufferInitDescriptor {
			label: Some("Camera Buffer"),
			contents: bytemuck::cast_slice(&[uniform]),
			usage: wgpu::BufferUsages::UNIFORM | wgpu::BufferUsages::COPY_DST,
		});

		let bind_group = device.create_bind_group(&wgpu::BindGroupDescriptor {
			layout: &layout,
			entries: &[wgpu::BindGroupEntry {
				binding: 0,
				resource: buffer.as_entire_binding(),
			}],
			label: Some("camera_bind_group"),
		});

		let info = CameraInfo {
			uniform,
			buffer,
			layout,
			bind_group,
		};

		(camera, info)
	}

	fn build_view_projection_matrix(&self) -> cgmath::Matrix4<f32> {
		let view = cgmath::Matrix4::look_at_rh(self.eye, self.target, self.up);
		let proj = cgmath::perspective(cgmath::Deg(self.fovy), 1.0, self.znear, self.zfar);
		OPENGL_TO_WGPU_MATRIX * proj * view
	}

	pub fn reset_camera(&mut self) {
		self.eye = Point3 {
			x: 0.0,
			y: 0.0,
			z: DEFAULT_CAMERA_ZOOM,
		};

		self.target = Point3 {
			x: 0.0,
			y: 0.0,
			z: 0.0,
		};
	}
}

#[repr(C)]
#[derive(Debug, Copy, Clone, bytemuck::Pod, bytemuck::Zeroable)]
pub struct CameraUniform {
	view_proj: [[f32; 4]; 4],
}

impl CameraUniform {
	fn new() -> Self {
		use cgmath::SquareMatrix;
		Self {
			view_proj: cgmath::Matrix4::identity().into(),
		}
	}

	pub fn update_view_proj(&mut self, camera: &Camera) {
		self.view_proj = camera.build_view_projection_matrix().into();
	}
}
