use std::sync::Arc;

use wgpu::util::DeviceExt;
use winit::{
	event::{ElementState, MouseScrollDelta, WindowEvent},
	keyboard::{Key, NamedKey},
};

use crate::controller::CameraController;

const REFRESH_RATE: f32 = 1.0 / 60.0;

pub struct Spotlight {
	window: Arc<winit::window::Window>,

	cursor_position: (f32, f32),
	size: f32,
	delta_size: f32,

	is_ctrl_pressed: bool,
}

pub struct SpotlightInfo {
	pub uniform: SpotlightUniform,
	pub buffer: wgpu::Buffer,
	pub layout: wgpu::BindGroupLayout,
	pub bind_group: wgpu::BindGroup,
}

impl Spotlight {
	pub fn new(device: &wgpu::Device, window: Arc<winit::window::Window>) -> (Self, SpotlightInfo) {
		let spotlight = Self {
			window,

			cursor_position: (960.0, 540.0),
			size: 50.0,
			delta_size: 0.0,

			is_ctrl_pressed: false,
		};

		let layout = device.create_bind_group_layout(&wgpu::BindGroupLayoutDescriptor {
			entries: &[wgpu::BindGroupLayoutEntry {
				binding: 0,
				visibility: wgpu::ShaderStages::FRAGMENT,
				ty: wgpu::BindingType::Buffer {
					ty: wgpu::BufferBindingType::Uniform,
					has_dynamic_offset: false,
					min_binding_size: None,
				},
				count: None,
			}],
			label: Some("spotlight_bind_group_layout"),
		});

		let uniform = SpotlightUniform {
			cursor_position: [0.0, 0.0],
			size: 50.0,
			zoom_level: 1.0,
			strength: 0.0,
			padding: 0.0,
		};

		let buffer = device.create_buffer_init(&wgpu::util::BufferInitDescriptor {
			label: Some("Spotlight Buffer"),
			contents: bytemuck::cast_slice(&[uniform]),
			usage: wgpu::BufferUsages::UNIFORM | wgpu::BufferUsages::COPY_DST,
		});

		let bind_group = device.create_bind_group(&wgpu::BindGroupDescriptor {
			layout: &layout,
			entries: &[wgpu::BindGroupEntry {
				binding: 0,
				resource: buffer.as_entire_binding(),
			}],
			label: Some("spotlight_bind_group"),
		});

		let info = SpotlightInfo {
			layout,
			bind_group,
			buffer,
			uniform,
		};

		(spotlight, info)
	}

	pub fn process_events(&mut self, event: &WindowEvent) {
		match event {
			WindowEvent::CursorMoved { position, .. } => {
				self.cursor_position = (position.x as f32, position.y as f32);
			}
			WindowEvent::KeyboardInput { event, .. } => {
				self.is_ctrl_pressed = false;
				let key = &event.logical_key;
				match event.state {
					ElementState::Pressed => {
						if let Key::Named(NamedKey::Control) = key {
							self.is_ctrl_pressed = true;
							self.window.set_cursor_visible(false);
						}
					}
					ElementState::Released => {
						if let Key::Named(NamedKey::Control) = key {
							self.is_ctrl_pressed = false;
							self.window.set_cursor_visible(true);
						}
					}
				}
			}
			WindowEvent::MouseWheel { delta, .. } => {
				if let MouseScrollDelta::LineDelta(_, y) = delta {
					if self.is_ctrl_pressed {
						self.delta_size += *y;
					}
				}
			}
			_ => {}
		}
	}

	pub fn update(&mut self) {
		if self.is_ctrl_pressed && self.delta_size.abs() > 0.01 {
			self.size = (self.size + self.delta_size * REFRESH_RATE).max(0.01);
			self.size = self.size.clamp(5.0, 250.0);
			self.delta_size -= self.delta_size * REFRESH_RATE * 0.1;
		}
	}
}

#[repr(C)]
#[derive(Debug, Copy, Clone, bytemuck::Pod, bytemuck::Zeroable)]
pub struct SpotlightUniform {
	cursor_position: [f32; 2],
	size: f32,
	zoom_level: f32,
	strength: f32,
	padding: f32,
}

impl SpotlightUniform {
	pub fn update(&mut self, spotlight: &Spotlight, controller: &CameraController) {
		self.cursor_position = [spotlight.cursor_position.0, spotlight.cursor_position.1];
		self.size = spotlight.size;
		self.zoom_level = controller.zoom_total.max(1.0);
		self.strength = if spotlight.is_ctrl_pressed { 0.9 } else { 0.0 };
	}
}
